{{ ir_reader_sony.spin

  Bob Belleville, 501d3r guy

  This object receives input from an IR remote which
  uses the Sony 12, 15 or 20 bit formats and places valid
  keycodes into a FiFo buffer.  Keycodes can then be removed
  and used in an application.  This buffer is sized below
  by the _qsize CON.  Unless the application goes off
  for a very long time without calling fifo_get, 8 or
  16 bytes is more than enough.

  A single cog is used to watch for input.  A cog, a pin,
  and a single lock bit must be available for this object
  to function.

  basic use is to:
    import this object
    init(pin, 0 or deviceID, repeatDelay, markRepeat)
    key := fifo_get
    if key == -1
      there isn't any new key
    else
      use key to do some thing

  see ir_reader_demo.spin for one example

  to modify for another kind of remote protocol:
    replace get_code/and perhaps get_pair
    
  see readme.pdf for more documentaton

  2007/03/02 - derived ir_reader_nec.spin
  2007/03/03 - various changes
  2011/04/08 - reworked to provide system clock frequency independent
               operation down to 10 MHz

}}
 
CON
        gapMin       =   62500  ' minimum idle gap for not auto-repeat
        startBitMin  =   2200   ' minimum length of start bit in us (2400 us reference)
        startBitMax  =   2600   ' maximum length of start bit in us (2400 us reference)
        oneBitMin    =   1000   ' minimum length of 1 (1200 us reference)
        oneBitMax    =   1400   ' maximum length of 1 (1200 us reference)
        maxBitTime   =   3000   ' maximum bit time (1800 us reference)

        ' Sony TV remote key codes
        ' these work for the remotes I tested however your mileage may vary
        one   =  0
        two   =  1
        three =  2
        four  =  3
        five  =  4
        six   =  5
        seven =  6
        eight =  7
        nine  =  8
        zero  =  9

        chUp  = 16
        chDn  = 17
        volUp = 18
        volDn = 19
        mute  = 20
        power = 21
        last  = 59

        _qsize  = 8             'must be a power of 2 (2,4,8,16,32,64,128,256 only)
        _qsm1   = _qsize-1      'mask

VAR

        long    lastvalid       'last valid code
        word    deviceID        'valid input device or
                                '  zero for any
                                '  get using ir_view for example
        long    lastcnt         'CNT at end last valid char
        byte    repeatlag       'ignore this many before
                                '  putting repeats in queue
        byte    repeatmark      'set high bit of repeated keycodes
                                '  if requested                                
        byte    irpin           'which pin to use

                                'fifo for key codes                                
        byte    lock            'hub lock index
        byte    head            'head (put) index for fifo
        byte    tail            'tail (get) index for fifo
        byte    fifo[_qsize]    'buffer itself

        long    gapMinCnt       'cnt value corresponding to gapMin
        long    startBitMinCnt  'cnt value corresponding to startBitMin
        long    oneBitMinCnt    'cnt value corresponding to oneBitMin
        long    maxBitTimeCnt   'cnt value corresponding to maxBitTime
        long    lastClkFreq     'last sampled value of clkfreq

        long    stack[24]       'for input cog


PUB init(pin, device, repeatdelay, markrepeat) | cog
{{
  pin         - port A input pin where IR receiver module is wired
  device      - 16 bit valid device code or 0 for any NEC device
  repeatdelay - number of repeated input codes to skip before
                adding last valid code to queue
  markrepeat  - true to set high bit of repeated key codes else
                not modified                

  returns cog number if running or -1 if failed              
}}
  irpin     := pin
  deviceID  := device
  repeatlag := repeatdelay
  ' set last clock frequency setting to zero to force initialization
  ' of timing parameters
  lastClkFreq~
  if markrepeat
    repeatmark := $80
  else
    repeatmark~
  lastcnt := cnt                'used to find repeat codes
  dira[irpin]~                  'input from pin (0 clear = input)
  lock := locknew               '!must have 1 lock available
  if lock == -1
    return -1
  fifo_flush                    'setup fifo
  lastvalid~                    'no last valid code
  cog := cognew(receive_ir,@stack)
  if cog == -1                  '!must have 1 cog available
    lockret(lock)               'clean up since cannot be used
    return -1
  return cog                    'all ok - off we go

PRI receive_ir | code, repeated
{{
  enqueue all valid keycodes
}}
  ' Initialize I/O within cog function
  dira[irpin]~                  'input from pin (0 clear = input)

  repeat
    code := get_code
                                'code is a repeat
    if code == $8000_0000 and lastvalid <> 0
      repeated++
      if repeated > repeatlag   'copy to queue after lag
        fifo_put((lastvalid>>16 & $FF) | repeatmark )
    else
      if deviceID               'accept only one device
        if deviceID <> code & $FFFF
          lastvalid~
          next                  'some other remote
      fifo_put(code>>16 & $FF)  'to buffer
      repeated~
      lastvalid := code

PRI get_code  | npairs, p, code, clkFreqMHz
{{
  wait for and return the next valid ir code
  return (in hex) either 00KKDDDD new code
                  or     80000000 repeat
  where DDDD is the 16 bit device ID
        KK   is the 8  bit key code                  
}}
  repeat
    ' if clock frequency has changed OR this is our first time
    ' being called, compute cnt values corresponding to
    ' timing parameters (all specified in microseconds)
    if lastClkFreq <> clkfreq
      lastClkFreq := clkfreq
      clkFreqMHz := clkfreq / 1_000_000
      gapMinCnt := gapMin * clkFreqMHz
      startBitMinCnt := startBitMin * clkFreqMHz
      oneBitMinCnt := oneBitMin * clkFreqMHz
      maxBitTimeCnt := maxBitTime * clkFreqMHz

    npairs~                     'count mark/space pairs
    code~                       'shift bits to code

    p := wait_start             'look for start condition

    if p < startBitMinCnt       'if mark time less than min start condition time,
      next                      '  glitch not a start code
    repeat                      'for each data and stop pair
      p := get_pair

      ' if there is actual mark time, we have not reached the end of the message
      if p > -1
        npairs++
        code >>= 1              'lsb sent first so shift right
        if p > oneBitMinCnt     'if mark time greater than min '1' bit time,
          code |= $8000_0000    '  shift in a 1
      else
        case npairs             'validate and  reformat
          12 :                  '12 bit format 7 key 5 device
            code := code>>4 & $7F0000 | code>>27 & $FFFF
          15 :                  '15 bit format 7 key 8 device
            code := code>>1 & $7F0000 | code>>24 & $FFFF
          20 :                  '15 bit format 7 key 11 device
            code := code<<4 & $7F0000 | code>>19 & $FFFF
          other:
            quit                'not a valid code try again

        'code is a repeat?
        if (cnt - lastcnt) < gapMinCnt
          lastcnt := cnt
          return $8000_0000

        lastcnt := cnt
        return code

PRI wait_start | irOnPulseWidth
{{
  wait for a potential start condition
}}
  irOnPulseWidth~               ' initialize return to 0

  ' setup counter A to keep track of low time
  frqa := 1
  ctra := (%10101 << 26) | (irpin)
  phsa~
  waitpeq(0, |< irpin, 0)
  waitpeq(|< irpin, |< irpin, 0)
  irOnPulseWidth := phsa

  return irOnPulseWidth

PRI get_pair | irOnPulseWidth, startCount
{{
  keep track of mark time (IR on)
  return mark time OR -1 if no mark time (no IR for max bit time)
}}
  irOnPulseWidth~~              'initialize return to -1 (no mark time)

  ' counter A already set up to keep track of low time
  ' frqa already set to 1
  ' clear phase accumulator A
  phsa~

  startCount := cnt

  ' wait here until our max zero bit time has expired or the ir pin has gone low
  ' NOTE: to handle cnt rollover, we follow the rule: (cnt - start > delay)
  repeat until ((ina[irpin] == 0) OR ((cnt - startCount) > maxBitTimeCnt))

  if (ina[irpin] == 0)
    waitpeq(|< irpin, |< irpin, 0)
    irOnPulseWidth := phsa

  return irOnPulseWidth

PUB get_lastvalid
''  return lastvalid keycode (may be 0 if not valid)
''  has format KKDDDD (key code & device code)
  return lastvalid

PUB fifo_flush
{{
  empty or initialize ir remote input
  first in first out queue
}}
  repeat while lockset(lock)
  head~
  tail~
  lockclr(lock)

PUB fifo_put(code) | len
{{
  if space is available insert code at head of
  fifo and return true
  else return false

  (left public but be careful this is usually
   only called by the input cog)
}}
  repeat while lockset(lock)
  len := head-tail              'needed to correct for 255->0
                                'tail could be at say 254 diff
                                'is neg and not the correct
                                'number of bytes in the queue
  if len < 0
    len += 256
  if len => _qsize
    lockclr(lock)
    return false
  fifo[head++ & _qsm1] := code
  lockclr(lock)
  return true

PUB fifo_get | code
{{
  return next available code
  or -1 if fifo empty
}}
  repeat while lockset(lock)
  if head == tail
    lockclr(lock)
    return -1
  code := fifo[tail++ & _qsm1]
  lockclr(lock)
  return code

PUB fifo_get_lastvalid
''  return lastvalid long to see device gode
  return lastvalid

PUB fifo_debug | ht
''  return fifo head and tail indexes in a long
  repeat while lockset(lock)
  ht := head<<16 | tail
  lockclr(lock)
  return ht
